#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.django.rattail.admin`` -- Admin Interface
"""

from django.contrib import admin
from django.contrib.admin.widgets import AdminTextInputWidget

from rattail.django.rattail.models import *
from rattail.django.rattail.models.rattail import GPCField


class ModelAdmin(admin.ModelAdmin):
    """
    Base admin manager for all Rattail models.
    """

    exclude = ('uuid',)


class ContactInfoInline(admin.TabularInline):
    """
    Base inline manager for all contact info models.
    """

    exclude = ('uuid', 'parent_type')
    extra = 0


class PartyInline(admin.TabularInline):
    """
    Base inline manager for all party models.
    """

    exclude = ('uuid', 'parent_type')
    extra = 0


class PersonPhoneNumberInline(ContactInfoInline):
    model = PersonPhoneNumber

class PersonEmailAddressInline(ContactInfoInline):
    model = PersonEmailAddress

class PersonPostalAddressInline(ContactInfoInline):
    model = PersonPostalAddress

class PersonAdmin(ModelAdmin):
    inlines = [PersonPhoneNumberInline, PersonEmailAddressInline,
               PersonPostalAddressInline]

    def get_formsets(self, request, obj=None):
        for inline in self.get_inline_instances(request):
            if isinstance(inline, ContactInfoInline) and obj is None:
                continue
            yield inline.get_formset(request, obj)

admin.site.register(Person, PersonAdmin)


admin.site.register(Role, ModelAdmin)


class UserRoleInline(admin.TabularInline):
    model = UserRole
    verbose_name = "Role"
    verbose_name_plural = "Roles"
    exclude = ('uuid',)
    extra = 0

class UserAdmin(ModelAdmin):
    exclude = ('uuid', 'password', 'salt')
    inlines = [UserRoleInline]

    def get_formsets(self, request, obj=None):
        for inline in self.get_inline_instances(request):
            if isinstance(inline, UserRoleInline) and obj is None:
                continue
            yield inline.get_formset(request, obj)

admin.site.register(User, UserAdmin)


class StorePhoneNumberInline(ContactInfoInline):
    model = StorePhoneNumber

class StoreEmailAddressInline(ContactInfoInline):
    model = StoreEmailAddress

class StorePostalAddressInline(ContactInfoInline):
    model = StorePostalAddress

class StoreAdmin(ModelAdmin):
    inlines = [StorePhoneNumberInline, StoreEmailAddressInline,
               StorePostalAddressInline]

    def get_formsets(self, request, obj=None):
        for inline in self.get_inline_instances(request):
            if isinstance(inline, ContactInfoInline) and obj is None:
                continue
            yield inline.get_formset(request, obj)

admin.site.register(Store, StoreAdmin)


admin.site.register(Department, ModelAdmin)


admin.site.register(Subdepartment, ModelAdmin)


admin.site.register(Category, ModelAdmin)


admin.site.register(Brand, ModelAdmin)


class VendorPhoneNumberInline(ContactInfoInline):
    model = VendorPhoneNumber

class VendorEmailAddressInline(ContactInfoInline):
    model = VendorEmailAddress

class VendorPostalAddressInline(ContactInfoInline):
    model = VendorPostalAddress

class VendorPartyInline(PartyInline):
    model = VendorParty

class VendorAdmin(ModelAdmin):
    inlines = [VendorPhoneNumberInline, VendorEmailAddressInline,
               VendorPostalAddressInline, VendorPartyInline]

    def get_formsets(self, request, obj=None):
        for inline in self.get_inline_instances(request):
            if isinstance(inline, ContactInfoInline) and obj is None:
                continue
            if isinstance(inline, PartyInline) and obj is None:
                continue
            yield inline.get_formset(request, obj)

admin.site.register(Vendor, VendorAdmin)


class ProductPriceInline(admin.TabularInline):
    model = ProductPrice
    verbose_name = "Price"
    verbose_name_plural = "Prices"
    exclude = ('uuid',)
    extra = 0

class ProductCostInline(admin.TabularInline):
    model = ProductCost
    verbose_name = "Cost"
    verbose_name_plural = "Costs"
    exclude = ('uuid',)
    extra = 0

class ProductAdmin(ModelAdmin):
    inlines = [ProductPriceInline, ProductCostInline]

    formfield_overrides = {
        GPCField: {
            'widget': AdminTextInputWidget,
            },
        }

    def get_form(self, request, obj=None, **kwargs):
        if obj is None:
            kwargs['exclude'] = ('uuid', 'regular_price', 'current_price')
        return super(ProductAdmin, self).get_form(request, obj, **kwargs)

    def get_formsets(self, request, obj=None):
        for inline in self.get_inline_instances(request):
            if isinstance(inline, ProductPriceInline) and obj is None:
                continue
            if isinstance(inline, ProductCostInline) and obj is None:
                continue
            yield inline.get_formset(request, obj)

admin.site.register(Product, ProductAdmin)


admin.site.register(Employee, ModelAdmin)


admin.site.register(CustomerGroup, ModelAdmin)


class CustomerPhoneNumberInline(ContactInfoInline):
    model = CustomerPhoneNumber

class CustomerEmailAddressInline(ContactInfoInline):
    model = CustomerEmailAddress

class CustomerPostalAddressInline(ContactInfoInline):
    model = CustomerPostalAddress

class CustomerGroupAssignmentInline(admin.TabularInline):
    model = CustomerGroupAssignment
    verbose_name = "Group Assignment"
    verbose_name_plural = "Group Assignments"
    exclude = ('uuid',)
    extra = 0

class CustomerPartyInline(PartyInline):
    model = CustomerParty

class CustomerAdmin(ModelAdmin):
    inlines = [CustomerPhoneNumberInline, CustomerEmailAddressInline,
               CustomerPostalAddressInline, CustomerGroupAssignmentInline,
               CustomerPartyInline]

    def get_formsets(self, request, obj=None):
        for inline in self.get_inline_instances(request):
            if isinstance(inline, ContactInfoInline) and obj is None:
                continue
            if isinstance(inline, CustomerGroupAssignmentInline) and obj is None:
                continue
            if isinstance(inline, PartyInline) and obj is None:
                continue
            yield inline.get_formset(request, obj)

admin.site.register(Customer, CustomerAdmin)


admin.site.register(LabelProfile, ModelAdmin)
