#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.django.rattail.models.contact`` -- Models for 'contact' Extension
"""

from __future__ import absolute_import

from django.db import models
from django.core.exceptions import ObjectDoesNotExist

from rattail.django.rattail.models import Model, uuid_field


__all__ = ['Contact', 'PhoneNumber', 'EmailAddress', 'PostalAddress',
           'Person', 'PersonPhoneNumber', 'PersonEmailAddress', 'PersonPostalAddress',
           'Party', 'PartyManager']


class ContactInfoManager(models.Manager):

    def __init__(self, parent_type):
        super(ContactInfoManager, self).__init__()
        self.parent_type = parent_type

    def get_query_set(self):
        qs = super(ContactInfoManager, self).get_query_set()
        qs = qs.filter(parent_type=self.parent_type)
        qs = qs.order_by('preference')
        return qs


class ContactInfo(Model):

    class Meta(Model.Meta):
        abstract = True

    uuid = uuid_field()
    parent_type = models.CharField(max_length=20)
    preference = models.IntegerField(blank=True, null=True)
    type = models.CharField(max_length=15, blank=True, null=True)


class PhoneNumber(ContactInfo):
    """
    Represents a phone (or fax) number associated with a contactable entity.
    """

    class Meta(ContactInfo.Meta):
        abstract = True
        db_table = ContactInfo.prefix('phone_numbers')
        verbose_name = "Phone Number"
        verbose_name_plural = "Phone Numbers"

    number = models.CharField(max_length=20)

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.number)

    def __unicode__(self):
        return unicode(self.number)


class EmailAddress(ContactInfo):
    """
    Represents an email address associated with a contactable entity.
    """

    class Meta(ContactInfo.Meta):
        abstract = True
        db_table = ContactInfo.prefix('email_addresses')
        verbose_name = "Email Address"
        verbose_name_plural = "Email Addresses"

    address = models.CharField(max_length=255)

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.address)

    def __unicode__(self):
        return unicode(self.address)


class PostalAddress(ContactInfo):
    """
    Represents a postal (mailing) address associated with a contactable entity.
    """

    class Meta(ContactInfo.Meta):
        abstract = True
        db_table = ContactInfo.prefix('postal_addresses')
        verbose_name = "Postal Address"
        verbose_name_plural = "Postal Addresses"

    street = models.CharField(max_length=100, blank=True, null=True)
    street2 = models.CharField(max_length=100, blank=True, null=True)
    city = models.CharField(max_length=100, blank=True, null=True)
    state = models.CharField(max_length=2, blank=True, null=True)
    zipcode = models.CharField(max_length=10, blank=True, null=True)

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, self.street)

    def __unicode__(self):
        return unicode(self.street)


class Contact(Model):
    """
    Base class for models which may be associated with contact info data.

    This class exists only to define convenience methods and properties on the
    derived class.
    """

    class Meta(Model.Meta):
        abstract = True

    def __init__(self, *args, **kwargs):
        super(Contact, self).__init__(*args, **kwargs)
        self._phones = {}
        self._emails = {}
        self._addresses = {}
        self._preferred = {}

    def _translate_info_class(self, class_):
        from rattail.django.rattail import models
        class_name = self.__class__.__name__ + class_.__name__
        if hasattr(models, class_name):
            class_ = getattr(models, class_name)
        return class_

    def _filter_queryset(self, qs):
        parent = self.__class__.__name__.lower()
        kwargs = {parent: self}
        return qs.filter(**kwargs)

    def _assign_parent(self, info):
        parent = self.__class__.__name__.lower()
        setattr(info, parent, self)

    def _get_info_dict(self, class_):
        if issubclass(class_, PhoneNumber):
            return self._phones
        elif issubclass(class_, EmailAddress):
            return self._emails
        elif issubclass(class_, PostalAddress):
            return self._addresses
        raise ValueError("Expected a subclass of ContactInfo; got %s" % class_.__name__)

    def _get_info_key(self, class_):
        if issubclass(class_, PhoneNumber):
            return 'phone'
        elif issubclass(class_, EmailAddress):
            return 'email'
        elif issubclass(class_, PostalAddress):
            return 'address'
        raise ValueError("Expected a subclass of ContactInfo; got %s" % class_.__name__)

    def _get_info(self, class_, type, info_dict=None):
        class_ = self._translate_info_class(class_)
        if info_dict is None:
            info_dict = self._get_info_dict(class_)
        if type not in info_dict:
            qs = class_.objects
            qs = self._filter_queryset(qs)
            qs = qs.filter(type=type)
            try:
                info = qs.get()
            except ObjectDoesNotExist:
                info = None
            info_dict[type] = info
        return info_dict[type]

    def get_phone(self, type):
        return self._get_info(PhoneNumber, type)

    def get_email(self, type):
        return self._get_info(EmailAddress, type)

    def get_address(self, type):
        return self._get_info(PostalAddress, type)

    def _update_info(self, class_, type, data):
        class_ = self._translate_info_class(class_)
        info_dict = self._get_info_dict(class_)
        info = self._get_info(class_, type, info_dict)

        # Check data for values; if none exist then we'll delete the info.
        values = False
        for value in data.itervalues():
            if value:
                values = True
                break

        if values:
            if not info:
                info = class_()
                self._assign_parent(info)
                info.type = type
                info_dict[type] = info
            for key, value in data.iteritems():
                setattr(info, key, value)
            info.save()
            return info

        if info:
            info.delete()
            info_dict[type] = None
        return None

    def update_phone(self, type, data):
        return self._update_info(PhoneNumber, type, data)

    def update_email(self, type, data):
        return self._update_info(EmailAddress, type, data)

    def update_address(self, type, data):
        return self._update_info(PostalAddress, type, data)

    def _get_preferred_info(self, class_):
        class_ = self._translate_info_class(class_)
        info_key = self._get_info_key(class_)
        if info_key not in self._preferred:
            qs = class_.objects
            qs = self._filter_queryset(qs)
            qs = qs.filter(preference=1)
            try:
                info = qs.get()
            except ObjectDoesNotExist:
                info = None
            self._preferred[info_key] = info
        return self._preferred[info_key]

    def get_preferred_phone(self):
        return self._get_preferred_info(PhoneNumber)

    def get_preferred_email(self):
        return self._get_preferred_info(EmailAddress)

    def get_preferred_address(self):
        return self._get_preferred_info(PostalAddress)

    @property
    def phone(self):
        return self.get_preferred_phone()

    @property
    def email(self):
        return self.get_preferred_email()

    @property
    def address(self):
        return self.get_preferred_address()

    def _set_preferred_info(self, class_, info):
        class_ = self._translate_info_class(class_)
        info_dict = self._get_info_dict(class_)
        info_key = self._get_info_key(class_)

        if info.preference != 1:
            info.preference = 1
            info.save()
        self._preferred[info_key] = info

        preference = 2
        for other_info in sorted(info_dict.values(),
                                key=lambda x: x.preference if x else 999):
            if other_info and other_info is not info:
                other_info.preference = preference
                other_info.save()
                preference += 1

    def set_preferred_phone(self, phone):
        self._set_preferred_info(PhoneNumber, phone)

    def set_preferred_email(self, email):
        self._set_preferred_info(EmailAddress, email)

    def set_preferred_address(self, address):
        self._set_preferred_info(PostalAddress, address)


class Person(Contact):
    """
    Represents a real, living and breathing person.  (Or, at least was
    previously living and breathing, in the case of the deceased.)
    """

    class Meta(Contact.Meta):
        db_table = Contact.prefix('people')
        verbose_name_plural = "People"

    uuid = uuid_field()
    first_name = models.CharField(max_length=50, blank=True, null=True)
    middle_name = models.CharField(max_length=50, blank=True, null=True)
    last_name = models.CharField(max_length=50, blank=True, null=True)
    display_name = models.CharField(max_length=100, blank=True, null=True, verbose_name="Display Name")
    driver_license_state = models.CharField(max_length=4, blank=True, null=True)
    driver_license_number = models.CharField(max_length=20, blank=True, null=True)

    def __repr__(self):
        return "<Person: %s>" % self.display_name

    def __unicode__(self):
        return unicode(self.display_name or '')

    def update_display_name(self):
        orig_display_name = self.display_name
        if self.first_name and self.last_name:
            self.display_name = self.first_name + ' ' + self.last_name
        elif self.first_name:
            self.display_name = self.first_name
        elif self.last_name:
            self.display_name = self.last_name
        return self.display_name != orig_display_name

    def save(self, *args, **kwargs):
        if not self.display_name:
            self.update_display_name()
        super(Person, self).save(*args, **kwargs)

    _employee = None

    @property
    def employee(self):
        if self._employee is None:
            from rattail.django.rattail.models import Employee
            qs = Employee.objects
            qs = qs.filter(person=self)
            try:
                self._employee = qs.get()
            except ObjectDoesNotExist:
                pass
        return self._employee

    def make_employee(self, **kwargs):
        if self.employee:
            return self.employee

        from rattail.django.rattail.models import Employee
        employee = Employee(**kwargs)
        employee.person = self
        employee.save()
        self._employee = employee
        return employee

    def unmake_employee(self):
        if self.employee:
            self.employee.delete()
            self._employee = None


class PersonContactInfo(models.Model):
    """
    Base class for person contact models (phone number, etc.).
    """

    class Meta:
        abstract = True

    person = models.ForeignKey(Person, db_column='parent_uuid')
    objects = ContactInfoManager('Person')

    def save(self, *args, **kwargs):
        self.parent_type = 'Person'
        super(PersonContactInfo, self).save(*args, **kwargs)


class PersonPhoneNumber(PhoneNumber, PersonContactInfo):
    """
    Represents a phone (or fax) number associated with a :class:`Person`.
    """


class PersonEmailAddress(EmailAddress, PersonContactInfo):
    """
    Represents an email address associated with a :class:`Person`.
    """    


class PersonPostalAddress(PostalAddress, PersonContactInfo):
    """
    Represents a postal (mailing) address associated with a :class:`Person`.
    """    


class PartyManager(models.Manager):

    def __init__(self, parent_type):
        super(PartyManager, self).__init__()
        self.parent_type = parent_type

    def get_query_set(self):
        qs = super(PartyManager, self).get_query_set()
        qs = qs.filter(parent_type=self.parent_type)
        qs = qs.order_by('preference')
        return qs


class Party(Model):
    """
    Base class for "party" relationships, e.g. sales contacts for vendors and
    authorized shoppers for customers.
    """

    class Meta(Model.Meta):
        abstract = True
        db_table = Model.prefix('parties')
        verbose_name = "Party"
        verbose_name_plural = "Parties"

    uuid = uuid_field()
    parent_type = models.CharField(max_length=20)
    person = models.ForeignKey(Person, db_column='person_uuid')
    type = models.CharField(max_length=25, blank=True, null=True)
    preference = models.IntegerField(blank=True, null=True)

    def __unicode__(self):
        return u'%s (%s)' % (self.person, self.type)
