#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.django.rattail.models.core`` -- Core Model Stuff
"""

from django.db import models
from django.conf import settings

import edbob


__all__ = ['Model', 'uuid_field', 'ProxyModel', 'PersonProxyModel']


class Model(models.Model):
    """
    Base class for all Rattail models.
    """

    class Meta:
        abstract = True
        app_label = 'rattail'
        managed = bool(getattr(settings, 'RATTAIL_DJANGO_MANAGED', True))

    @classmethod
    def prefix(cls, table_name):
        """
        Returns a possibly-prefixed version of ``table_name``.

        The name returned will depend on whether or not a prefix is necessary,
        based on whether the Rattail tables are held in the native Django
        database, or an externally-managed one.
        """

        managed = bool(getattr(settings, 'RATTAIL_DJANGO_MANAGED', True))
        if managed:
            table_name = 'rattail_' + table_name
        return table_name


class ProxyModel(models.Model):
    """
    Base class for models which intend to "proxy" certain attributes to an
    underlying Rattail model.  This is designed for use in apps which use
    Rattail as their primary schema but need to extend it in some way.
    """

    class Meta:
        abstract = True

    _proxy_attribute_map = {}

    def __getattr__(self, name):
        for proxy, attrs in self.__class__._proxy_attribute_map.iteritems():
            for attr in attrs:
                if isinstance(attr, tuple):
                    convenient, actual = attr
                else:
                    convenient = actual = attr
                if convenient == name:
                    return getattr(getattr(self, proxy), actual)
        return super(ProxyModel, self).__getattr__(name)

    def __setattr__(self, name, value):
        for proxy, attrs in self.__class__._proxy_attribute_map.iteritems():
            for attr in attrs:
                if isinstance(attr, tuple):
                    convenient, actual = attr
                else:
                    convenient = actual = attr
                if convenient == name:
                    print 'setting %s.%s.%s (via %s) to %s' % (
                        self.__class__.__name__, proxy, actual, convenient, repr(value))
                    return setattr(getattr(self, proxy), actual, value)
        return super(ProxyModel, self).__setattr__(name, value)


class PersonProxyModel(ProxyModel):
    """
    Proxy model specific to :class:`rattail.django.rattail.models.Person`.
    """

    class Meta(ProxyModel.Meta):
        abstract = True

    _proxy_person = 'person'

    def __init__(self, *args, **kwargs):
        self._proxy_attribute_map[self._proxy_person] = [
            'first_name',
            'middle_name',
            'last_name',
            'driver_license_number',
            'driver_license_state',
            'address',
            'update_address',
            'set_preferred_address',
            'phone',
            'get_phone',
            'update_phone',
            'set_preferred_phone',
            'email',
            'update_email',
            'set_preferred_email',
            'make_employee',
            'unmake_employee',
            ]
        super(PersonProxyModel, self).__init__(*args, **kwargs)

    def save(self, **kwargs):
        getattr(self, self._proxy_person).save(**kwargs)
        super(PersonProxyModel, self).save(**kwargs)


class UUIDField(models.CharField):
    """
    "Custom" type for UUID fields.  This is actually only necessary so that the
    Django Admin site will include the UUID as a hidden field when rendering
    inline forms.

    See also `this ticket <https://code.djangoproject.com/ticket/13696>`_.
    """

    def contribute_to_class(self, cls, name):
        super(UUIDField, self).contribute_to_class(cls, name)
        cls._meta.has_auto_field = True
        cls._meta.auto_field = self


def uuid_field():
    """
    Convenience function which returns a ``uuid`` column for use as a table's
    primary key.
    """

    return UUIDField(max_length=32, default=edbob.get_uuid,
                     primary_key=True, verbose_name="UUID")
