#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``rattail.django.rattail.models.rattail`` -- Models for 'rattail' Extension
"""

from __future__ import absolute_import

from django.db import models
from django.core.exceptions import ObjectDoesNotExist

from edbob import EMAIL_PREFERENCE

import rattail
from rattail.gpc import GPC

from rattail.django.rattail.models import Model, uuid_field
from rattail.django.rattail.models.contact import (
    ContactInfoManager, PhoneNumber, EmailAddress, PostalAddress,
    PartyManager, Party, Person)
from rattail.django.util import enum_to_choices


__all__ = ['Store', 'StorePhoneNumber', 'StoreEmailAddress', 'StorePostalAddress',
           'Department', 'Subdepartment', 'Category', 'Brand',
           'Vendor', 'VendorPhoneNumber', 'VendorEmailAddress', 'VendorPostalAddress', 'VendorParty',
           'Product', 'ProductPrice', 'ProductCost',
           'Employee',
           'Customer', 'CustomerPhoneNumber', 'CustomerEmailAddress', 'CustomerPostalAddress',
           'CustomerGroup', 'CustomerGroupAssignment', 'CustomerParty',
           'LabelProfile']


class GPCField(models.BigIntegerField):
    """
    Custom type for the ``Product.upc`` field.
    """

    description = "Global Product Classification (GPC)"

    __metaclass__ = models.SubfieldBase

    def to_python(self, value):
        if value is None:
            return None
        return GPC(value)

    def get_prep_value(self, value):
        if value is None:
            return None
        return int(value)


class Store(Model):
    """
    Represents a store (physical or otherwise) within the organization.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('stores')
        verbose_name = "Store"

    uuid = uuid_field()
    id = models.CharField(max_length=10, blank=True, null=True, verbose_name="ID")
    name = models.CharField(max_length=100, blank=True, null=True)

    def __repr__(self):
        return "<Store: %s, %s>" % (self.id, self.name)

    def __unicode__(self):
        return unicode(self.name or '')


class StoreContactInfo(models.Model):
    """
    Base class for store contact models (phone number, etc.).
    """

    class Meta:
        abstract = True

    store = models.ForeignKey(Store, db_column='parent_uuid')
    objects = ContactInfoManager('Store')

    def save(self, *args, **kwargs):
        self.parent_type = 'Store'
        super(StoreContactInfo, self).save(*args, **kwargs)


class StorePhoneNumber(PhoneNumber, StoreContactInfo):
    """
    Represents a phone (or fax) number associated with a :class:`Store`.
    """


class StoreEmailAddress(EmailAddress, StoreContactInfo):
    """
    Represents an email address associated with a :class:`Store`.
    """


class StorePostalAddress(PostalAddress, StoreContactInfo):
    """
    Represents a postal (mailing) address associated with a :class:`Store`.
    """    


class Department(Model):
    """
    Represents an organizational department.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('departments')
        verbose_name = "Department"

    uuid = uuid_field()
    number = models.IntegerField(blank=True, null=True)
    name = models.CharField(max_length=30, blank=True, null=True)

    def __repr__(self):
        return "<Department: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name or '')


class Subdepartment(Model):
    """
    Represents an organizational subdepartment.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('subdepartments')
        verbose_name = "Subdepartment"

    uuid = uuid_field()
    number = models.IntegerField(blank=True, null=True)
    name = models.CharField(max_length=30, blank=True, null=True)
    department = models.ForeignKey(Department, db_column='department_uuid', blank=True, null=True)

    def __repr__(self):
        return "<Subdepartment: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name or '')


class Category(Model):
    """
    Represents an organizational category for products.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('categories')
        verbose_name = "Category"
        verbose_name_plural = "Categories"

    uuid = uuid_field()
    number = models.IntegerField(blank=True, null=True)
    name = models.CharField(max_length=50, blank=True, null=True)
    department = models.ForeignKey(Department, db_column='department_uuid', blank=True, null=True)

    def __repr__(self):
        return "<Category: %s, %s>" % (self.number, self.name)

    def __unicode__(self):
        return unicode(self.name or '')


class Brand(Model):
    """
    Represents a brand or similar product line.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('brands')
        verbose_name = "Brand"

    uuid = uuid_field()
    name = models.CharField(max_length=100, blank=True, null=True)

    def __repr__(self):
        return "<Brand: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name or '')


class Vendor(Model):
    """
    Represents a vendor from which products are purchased by the store.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('vendors')
        verbose_name = "Vendor"

    uuid = uuid_field()
    id = models.CharField(max_length=15, blank=True, null=True, verbose_name="ID")
    name = models.CharField(max_length=40, blank=True, null=True)
    special_discount = models.DecimalField(max_digits=5, decimal_places=3, blank=True, null=True, verbose_name="Special Discount")

    def __repr__(self):
        return "<Vendor: %s>" % self.name

    def __unicode__(self):
        return unicode(self.name or '')


class VendorContactInfo(models.Model):
    """
    Base class for vendor contact models (phone number, etc.).
    """

    class Meta:
        abstract = True

    vendor = models.ForeignKey(Vendor, db_column='parent_uuid')
    objects = ContactInfoManager('Vendor')

    def save(self, *args, **kwargs):
        self.parent_type = 'Vendor'
        super(VendorContactInfo, self).save(*args, **kwargs)


class VendorPhoneNumber(PhoneNumber, VendorContactInfo):
    """
    Represents a phone (or fax) number associated with a :class:`Vendor`.
    """


class VendorEmailAddress(EmailAddress, VendorContactInfo):
    """
    Represents an email address associated with a :class:`Vendor`.
    """


class VendorPostalAddress(PostalAddress, VendorContactInfo):
    """
    Represents a postal (mailing) address associated with a :class:`Vendor`.
    """    


class VendorParty(Party):
    """
    Represents a "party" (e.g. sales contact) for a vendor, from the retailer's
    perspective.
    """

    vendor = models.ForeignKey(Vendor, db_column='parent_uuid')
    objects = PartyManager('Vendor')

    def __repr__(self):
        return "<VendorParty: %s, %s>" % (self.vendor, self.person)

    def save(self, *args, **kwargs):
        self.parent_type = 'Vendor'
        super(VendorParty, self).save(*args, **kwargs)


class Product(Model):
    """
    Represents a product for sale and/or purchase.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('products')
        verbose_name = "Product"

    UNIT_OF_MEASURE_CHOICES = enum_to_choices(rattail.UNIT_OF_MEASURE)

    uuid = uuid_field()
    upc = GPCField(db_index=True, verbose_name="UPC")
    department = models.ForeignKey(Department, db_column='department_uuid', blank=True, null=True)
    subdepartment = models.ForeignKey(Subdepartment, db_column='subdepartment_uuid', blank=True, null=True)
    category = models.ForeignKey(Category, db_column='category_uuid', blank=True, null=True)
    brand = models.ForeignKey(Brand, db_column='brand_uuid', blank=True, null=True)
    description = models.CharField(max_length=60, blank=True, null=True)
    description2 = models.CharField(max_length=60, blank=True, null=True, verbose_name="Description Line 2")
    size = models.CharField(max_length=30, blank=True, null=True)
    unit_of_measure = models.CharField(max_length=4, choices=UNIT_OF_MEASURE_CHOICES, blank=True, null=True, verbose_name="Unit of Measure")
    regular_price = models.ForeignKey('ProductPrice', db_column='regular_price_uuid', related_name='+', blank=True, null=True, verbose_name="Regular Price")
    current_price = models.ForeignKey('ProductPrice', db_column='current_price_uuid', related_name='+', blank=True, null=True, verbose_name="Current Price")

    def __repr__(self):
        return "<Product: %s>" % self.description

    def __unicode__(self):
        return unicode(self.description or '')


class ProductPrice(Model):
    """
    Represents a price for a product.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('product_prices')
        verbose_name = "Product Price"

    PRICE_TYPE_CHOICES = enum_to_choices(rattail.PRICE_TYPE)

    uuid = uuid_field()
    product = models.ForeignKey(Product, db_column='product_uuid')
    type = models.IntegerField(choices=PRICE_TYPE_CHOICES, blank=True, null=True)
    level = models.IntegerField(blank=True, null=True)
    starts = models.DateTimeField(blank=True, null=True)
    ends = models.DateTimeField(blank=True, null=True)
    price = models.DecimalField(max_digits=8, decimal_places=3, blank=True, null=True)
    multiple = models.IntegerField(blank=True, null=True)
    pack_price = models.DecimalField(max_digits=8, decimal_places=3, blank=True, null=True)
    pack_multiple = models.IntegerField(blank=True, null=True)

    def __repr__(self):
        return "<ProductPrice: %s : %s>" % (self.product, self.price)

    def __unicode__(self):
        if self.price is None:
            return u''
        type = rattail.PRICE_TYPE.get(self.type, "?")
        return u'$ %0.2f / %d (%s)' % (self.price, self.multiple or 1, type)


class ProductCost(Model):
    """
    Represents a source from which a product may be obtained via purchase.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('product_costs')
        verbose_name = "Product Cost"

    uuid = uuid_field()
    product = models.ForeignKey(Product, db_column='product_uuid')
    vendor = models.ForeignKey(Vendor, db_column='vendor_uuid')
    preference = models.IntegerField()
    code = models.CharField(max_length=15, blank=True, null=True)
    case_size = models.IntegerField(blank=True, null=True, verbose_name="Case Size")
    case_cost = models.DecimalField(max_digits=9, decimal_places=5, blank=True, null=True, verbose_name="Case Cost")
    pack_size = models.IntegerField(blank=True, null=True, verbose_name="Pack Size")
    pack_cost = models.DecimalField(max_digits=9, decimal_places=5, blank=True, null=True, verbose_name="Pack Cost")
    unit_cost = models.DecimalField(max_digits=9, decimal_places=5, blank=True, null=True, verbose_name="Unit Cost")
    effective = models.DateTimeField(blank=True, null=True)

    def __repr__(self):
        return "<ProductCost: %s : %s>" % (self.product, self.vendor)

    def __unicode__(self):
        if self.unit_cost is None:
            return u''
        return '$ %0.4f (%s)' % (self.unit_cost, self.vendor)


class Employee(Model):
    """
    Represents an employee within the organization.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('employees')
        verbose_name = "Employee"

    EMPLOYEE_STATUS_CHOICES = enum_to_choices(rattail.EMPLOYEE_STATUS)

    uuid = uuid_field()
    id = models.CharField(max_length=20, blank=True, null=True, verbose_name="ID")
    person = models.OneToOneField(Person, db_column='person_uuid', related_name='+')
    status = models.IntegerField(choices=EMPLOYEE_STATUS_CHOICES, blank=True, null=True)
    display_name = models.CharField(max_length=100, blank=True, null=True, verbose_name="Display Name")

    def __repr__(self):
        return "<Employee: %s>" % self.person

    def __unicode__(self):
        return unicode(self.display_name or self.person)

    def save(self, *args, **kwargs):
        if not self.display_name and self.person:
            if self.person.display_name:
                self.display_name = self.person.display_name
            elif self.person.first_name and self.person.last_name:
                self.display_name = self.person.first_name + ' ' + self.person.last_name
            elif self.person.first_name:
                self.display_name = self.person.first_name
            elif self.person.last_name:
                self.display_name = self.person.last_name
        super(Employee, self).save(*args, **kwargs)


class Customer(Model):
    """
    Represents a customer account.  Customer accounts may consist of more than
    one :class:`Person`, in some cases.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('customers')
        verbose_name = "Customer"

    EMAIL_PREFERENCE_CHOICES = enum_to_choices(EMAIL_PREFERENCE)

    uuid = uuid_field()
    id = models.CharField(max_length=20, null=True, blank=True, verbose_name="ID")
    name = models.CharField(max_length=255, null=True, blank=True)
    email_preference = models.IntegerField(choices=EMAIL_PREFERENCE_CHOICES, null=True, blank=True, verbose_name="Email Preference")

    def __repr__(self):
        return "<Customer: %s, %s>" % (self.id, self.name)

    def __unicode__(self):
        return unicode(self.name or '')

    _primary_party = None

    @property
    def primary_party(self):
        if self._primary_party is None:
            qs = CustomerParty.objects
            qs = qs.filter(customer=self)
            qs = qs.filter(preference=1)
            try:
                self._primary_party = qs.get()
            except ObjectDoesNotExist:
                pass
        return self._primary_party

    def update_name(self):
        orig_name = self.name
        if self.primary_party:
            person = self.primary_party.person
            if not person.display_name:
                person.update_display_name()
            self.name = person.display_name
        return self.name != orig_name

    def save(self, *args, **kwargs):
        if not self.name:
            self.update_name()
        super(Customer, self).save(*args, **kwargs)


class CustomerContactInfo(models.Model):
    """
    Base class for customer contact models (phone number, etc.).
    """

    class Meta:
        abstract = True

    customer = models.ForeignKey(Customer, db_column='parent_uuid')
    objects = ContactInfoManager('Customer')

    def save(self, *args, **kwargs):
        self.parent_type = 'Customer'
        super(CustomerContactInfo, self).save(*args, **kwargs)


class CustomerPhoneNumber(PhoneNumber, CustomerContactInfo):
    """
    Represents a phone (or fax) number associated with a :class:`Customer`.
    """


class CustomerEmailAddress(EmailAddress, CustomerContactInfo):
    """
    Represents an email address associated with a :class:`Customer`.
    """


class CustomerPostalAddress(PostalAddress, CustomerContactInfo):
    """
    Represents a postal (mailing) address associated with a :class:`Customer`.
    """    


class CustomerParty(Party):
    """
    Represents a "party" (e.g. authorized shopper) for a customer, from the
    retailer's perspective.
    """

    customer = models.ForeignKey(Customer, db_column='parent_uuid')
    objects = PartyManager('Customer')

    def __repr__(self):
        return "<CustomerParty: %s, %s>" % (self.customer, self.person)

    def save(self, *args, **kwargs):
        self.parent_type = 'Customer'
        super(CustomerParty, self).save(*args, **kwargs)


class CustomerGroup(Model):
    """
    Represents an arbitrary group to which :class:`Customer` instances may (or
    may not) belong.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('customer_groups')
        verbose_name = "Customer Group"

    uuid = uuid_field()
    id = models.CharField(max_length=20, blank=True, null=True, verbose_name="ID")
    name = models.CharField(max_length=255, blank=True, null=True)

    def __repr__(self):
        return "<CustomerGroup: %s, %s>" % (self.id, self.name)

    def __unicode__(self):
        return unicode(self.name or '')


class CustomerGroupAssignment(Model):
    """
    Represents the assignment of a :class:`Customer` to a
    :class:`CustomerGroup`.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('customers_groups')
        verbose_name = "Customer Group Assignment"

    uuid = uuid_field()
    customer = models.ForeignKey(Customer, db_column='customer_uuid')
    group = models.ForeignKey(CustomerGroup, db_column='group_uuid')
    ordinal = models.IntegerField()

    def __repr__(self):
        return "<CustomerGroupAssignment: %s, %s>" % (self.customer, self.group)

    def __unicode__(self):
        return u'%s, %s' % (self.customer, self.group)


class LabelProfile(Model):
    """
    Represents a "profile" (collection of settings) for product label printing.
    """

    class Meta(Model.Meta):
        db_table = Model.prefix('label_profiles')
        verbose_name = "Label Profile"

    uuid = uuid_field()
    ordinal = models.IntegerField(blank=True, null=True)
    code = models.CharField(max_length=3, blank=True, null=True)
    description = models.CharField(max_length=50, blank=True, null=True)
    printer_spec = models.CharField(max_length=255, blank=True, null=True, verbose_name="Printer Spec")
    formatter_spec = models.CharField(max_length=255, blank=True, null=True, verbose_name="Formatter Spec")
    format = models.TextField(blank=True, null=True)
    visible = models.BooleanField()

    def __repr__(self):
        return "<LabelProfile: %s>" % self.code

    def __unicode__(self):
        return unicode(self.description or '')
